{
  MapEdit 4.1     Wolfenstein Map Editor

     Copyright (c) 1992  Bill Kirby

  ver 4.1a (Dave Huntoon)
      - Adds ability to open Spear of Destiny (SOD) maps.
      - Allows access to objects > 00FE.  Needed for SOD
	objects.
      - minor fix to completely clear text area below
	the map display when the mouse is moved outisde
	of the map area.
  ver 4.2 (Petri Hrkk) 3.10.1993
	    - command E will export current level to transport file
	      if press just enter to file name so none will export.
	      Correct file extensions are '.WL1', '.WL3', '.WL6' and
	      '.SOD'. Map filename is ????MAP.??? and head filename
	      is ????HEAD.???
	    - command I will inport transport file to current level
	      if file not found so level is so that it was

}

{$A+,B-,D-,E-,F-,G-,I+,L-,N-,O-,R-,S-,V-,X-}
{$M 16384,0,655360}
program mapedit;

uses crt,dos,graph,mouse;

const
      MAP_X = 6;
      MAP_Y = 6;
      TEXTLOC = 460;

      GAMEPATH     : string = '.\';
      HEADFILENAME : string = 'maphead';
      MAPFILENAME  : string = 'maptemp';
      MAPFNAME     : string = 'ransmap';
      HEADFNAME    : string = 'ranshead';
      EXTT         : string = '.WL1';
      LEVELS       : word   = 10;
      GAME_VERSION : real   = 1.0;

type data_block = record
       size : word;
       data : pointer;
     end;

     level_type = record
       map,
       objects,
       other           : data_block;
       width,
       height          : word;
       name            : string[16];
     end;

     grid = array[0..63,0..63] of word;

     filltype = (solid,check);
     doortype = (horiz,vert);


var levelmap,
    objectmap    : grid;
    maps         : array[1..60] of level_type;

    show_objects,
    show_floor   : Boolean;

    mapgraph,
    objgraph     : array[0..511] of string[4];
    mapnames,
    objnames     : array[0..511] of string[20];

    themouse  : resetrec;
    mouseloc  : locrec;

function upper(s: string): string;
var i: integer;
begin
  for i:=1 to Length(s) do
    if s[i] in ['a'..'z'] then
      s[i]:=Chr(Ord(s[i])-Ord('a')+Ord('A'));
  upper:=s;
end;

procedure decorate(x,y,c: integer);
var i,j: integer;
begin
  SetFillStyle(1,c);
  Bar(x*7+MAP_X+2,y*7+MAP_Y+2,x*7+MAP_X+4,y*7+MAP_Y+4);
end;

procedure box(fill: filltype; x,y,c1,c2: integer; dek: Boolean);
begin
  if fill=solid then
    SetFillStyle(1,c1)
  else
    SetFillStyle(9,c1);

  Bar(x*7+MAP_X,y*7+MAP_Y,x*7+6+MAP_X,y*7+6+MAP_Y);
  if dek then decorate(x,y,c2);
end;

procedure OutText(x,y,color: integer; s: string);
begin
  SetColor(color);
  OutTextXY(x*7+MAP_X,y*7+MAP_Y,s);
end;

function hex(x: word): string;
const digit : string[16] = '0123456789ABCDEF';
var temp : string[4];
    i    : integer;
begin
  temp:= '    ';
  for i:= 4 downto 1 do
    begin
      temp[i]:= digit[(x and $000f)+1];
      x:= x div 16;
    end;
  hex:= temp;
end;

function hexbyte(x: byte): string;
const digit : string[16] = '0123456789ABCDEF';
var temp : string[4];
    i    : integer;
begin
  temp:= '  ';
  for i:= 2 downto 1 do
    begin
      temp[i]:= digit[(x and $000f)+1];
      x:= x div 16;
    end;
  hexbyte:= temp;
end;

procedure doline(x,y,x2,y2: integer);
begin
  Line(x+MAP_X,y+MAP_Y,x2+MAP_X,y2+MAP_Y);
end;

procedure dobar(x,y,x2,y2: integer);
begin
  Bar(x+MAP_Y,y+MAP_Y,x2+MAP_X,y2+MAP_Y);
end;

procedure circle(x,y,c1,c2: integer);
const sprite : array[0..6,0..6] of byte =
       ((0,0,1,1,1,0,0),
	(0,1,1,1,1,1,0),
	(1,1,1,2,1,1,1),
	(1,1,2,2,2,1,1),
	(1,1,1,2,1,1,1),
	(0,1,1,1,1,1,0),
	(0,0,1,1,1,0,0));
var i,j,c: integer;
begin
  for i:= 0 to 6 do
    for j:= 0 to 6 do
      begin
  case sprite[i,j] of
    0: c:=0;
    1: c:=c1;
    2: c:=c2;
  end;
	PutPixel(x*7+MAP_X+i,y*7+MAP_Y+j,c);
      end;
end;

procedure door(dtype: doortype; x,y,color: integer);
begin
  case dtype of
    vert: begin
	    SetFillStyle(1,color);
      dobar(x*7+2,y*7,x*7+4,y*7+6);
    end;
    horiz : begin
	      SetFillStyle(1,color);
	dobar(x*7,y*7+2,x*7+6,y*7+4);
    end;
  end;
end;

function hexnibble(c: char): byte;
begin
  case c of
    '0'..'9': hexnibble:= Ord(c)-Ord('0');
    'a'..'f': hexnibble:= Ord(c)-Ord('a')+10;
    'A'..'F': hexnibble:= Ord(c)-Ord('A')+10;
    else hexnibble:= 0;
  end;
end;

procedure output(x,y: integer; data: string);
var size  : integer;
    temp  : string[4];
    c1,c2 : byte;
begin
  if data<>'0000' then
    begin
      temp:= data;
      c1:= hexnibble(temp[1]);
      c2:= hexnibble(temp[2]);
      case temp[3] of
	'0': OutText(x,y,c1,temp[4]);
  '1': box(solid,x,y,c1,c2,false);
  '2': box(check,x,y,c1,c2,false);
  '3': box(solid,x,y,c1,c2,true);
  '4': box(check,x,y,c1,c2,true);
  '5': circle(x,y,c1,c2);
  '6': door(horiz,x,y,c1);
  '7': door(vert,x,y,c1);
  '8': begin
	       SetFillStyle(1,c1);
	 dobar(x*7,y*7,x*7+6,y*7+3);
	       SetFillStyle(1,c2);
	 dobar(x*7,y*7+4,x*7+6,y*7+6);
	end;
	'9': PutPixel(x*7+MAP_X+3,y*7+MAP_Y+3,c1);
	'a': begin SetFillStyle(1,c1); dobar(x*7+2,y*7+1,x*7+4,y*7+5); end;
	'b': begin SetFillStyle(1,c1); dobar(x*7+2,y*7+2,x*7+4,y*7+4); end;
	'c': begin SetFillStyle(1,c1); dobar(x*7+1,y*7+1,x*7+5,y*7+5); end;
  'd': begin
	       SetColor(c1);
	 doline(x*7+1,y*7+1,x*7+5,y*7+5);
	 doline(x*7+5,y*7+1,x*7+1,y*7+5);
       end;
  'e': begin
	       SetColor(c1);
	       Rectangle(x*7+MAP_X,y*7+MAP_Y,x*7+MAP_X+6,y*7+MAP_Y+6);
       end;
  'f': case c2 of
	2: begin {east}
		   SetColor(c1);
       doline(x*7,y*7+3,x*7+6,y*7+3);
       doline(x*7+6,y*7+3,x*7+3,y*7);
		   doline(x*7+6,y*7+3,x*7+3,y*7+6);
    end;
	0: begin {north}
		   SetColor(c1);
       doline(x*7+3,y*7+6,x*7+3,y*7);
       doline(x*7+3,y*7,x*7,y*7+3);
       doline(x*7+3,y*7,x*7+6,y*7+3);
     end;
	6: begin {west}
		   SetColor(c1);
       doline(x*7+6,y*7+3,x*7,y*7+3);
       doline(x*7,y*7+3,x*7+3,y*7);
       doline(x*7,y*7+3,x*7+3,y*7+6);
     end;
	4: begin {south}
		   SetColor(c1);
       doline(x*7+3,y*7,x*7+3,y*7+6);
       doline(x*7+3,y*7+6,x*7,y*7+3);
       doline(x*7+3,y*7+6,x*7+6,y*7+3);
     end;
	1: begin {northeast}
		   SetColor(c1);
       doline(x*7,y*7+6,x*7+6,y*7);
       doline(x*7+6,y*7,x*7+3,y*7);
       doline(x*7+6,y*7,x*7+6,y*7+3);
     end;
	7: begin {northwest}
		   SetColor(c1);
       doline(x*7+6,y*7+6,x*7,y*7);
       doline(x*7,y*7,x*7+3,y*7);
       doline(x*7,y*7,x*7,y*7+3);
     end;
	3: begin {southeast}
		   SetColor(c1);
		   doline(x*7,y*7,x*7+6,y*7+6);
       doline(x*7+6,y*7+6,x*7+3,y*7+6);
       doline(x*7+6,y*7+6,x*7+6,y*7+3);
     end;
	5: begin {southwest}
		   SetColor(c1);
       doline(x*7+6,y*7,x*7,y*7+6);
       doline(x*7,y*7+6,x*7+3,y*7+6);
       doline(x*7,y*7+6,x*7,y*7+3);
     end;

       end;
      end;
    end;
end;

procedure display_map;
var i,j: integer;
begin
  j:= 63;
  i:= 0;
  repeat
    SetFillStyle(1,0);
    dobar(i*7,j*7,i*7+6,j*7+6);
    if show_floor then
      output(i,j,mapgraph[levelmap[i,j]])
    else
      if not (levelmap[i,j] in [$6a..$8f]) then
  output(i,j,mapgraph[levelmap[i,j]]);
    if show_objects then
      output(i,j,objgraph[objectmap[i,j]]);
    Inc(i);
    if i=64 then
      begin
	i:= 0;
	Dec(j);
      end;
  until (j<0) or KeyPressed;
end;

procedure read_levels;
var headfile,
    mapfile  : file;
    outstr   : string;
    s,o,
    size     : word;
    idsig    : string[4];
    level    : integer;
    levelptr : longint;
    tempstr  : string[16];
    key                      : char;
    map_pointer,
    object_pointer,
    other_pointer    : longint;

begin
  idsig:= '    ';
  tempstr:= '                ';
  Assign(headfile,GAMEPATH+HEADFILENAME);
  {$I-}
  Reset(headfile,1);
  {$I+}
  if ioresult<>0 then
    begin
      WriteLn('error opening ',HEADFILENAME);
      Halt(1);
    end;
  Assign(mapfile,GAMEPATH+MAPFILENAME);
  {$I-}
  Reset(mapfile,1);
  {$I+}
  if ioresult<>0 then
    begin
      WriteLn('error opening ',MAPFILENAME);
      Halt(1);
    end;

  for level:= 1 to LEVELS do
    begin
      Seek(headfile,2+(level-1)*4);
      BlockRead(headfile,levelptr,4);
      Seek(mapfile,levelptr);
      with maps[level] do
  begin
	  BlockRead(mapfile,map_pointer,4);
	  BlockRead(mapfile,object_pointer,4);
	  BlockRead(mapfile,other_pointer,4);
	  BlockRead(mapfile,map.size,2);
	  BlockRead(mapfile,objects.size,2);
	  BlockRead(mapfile,other.size,2);
	  BlockRead(mapfile,width,2);
	  BlockRead(mapfile,height,2);
	  name[0]:=#16;
	  BlockRead(mapfile,name[1],16);
	  if GAME_VERSION = 1.1 then
	    BlockRead(mapfile,idsig[1],4);
	  Seek(mapfile,map_pointer);
	  GetMem(map.data,map.size);
	  s:= Seg(map.data^);
	  o:= Ofs(map.data^);
	  BlockRead(mapfile,Mem[s:o],map.size);

	  Seek(mapfile,object_pointer);
	  GetMem(objects.data,objects.size);
	  s:= Seg(objects.data^);
	  o:= Ofs(objects.data^);
	  BlockRead(mapfile,Mem[s:o],objects.size);

    Seek(mapfile,other_pointer);
	  GetMem(other.data,other.size);
	  s:= Seg(other.data^);
	  o:= Ofs(other.data^);
	  BlockRead(mapfile,Mem[s:o],other.size);
	  if GAME_VERSION = 1.0 then
	    BlockRead(mapfile,idsig[1],4);
	end;
    end;
  Close(mapfile);
  Close(headfile);
end;

procedure write_levels;
var headfile,
    mapfile    : file;
    abcd,
    s,o,
    size     : word;
    idsig    : string[4];
    level    : integer;
    levelptr : longint;
    tempstr  : string[16];
    map_pointer,
    object_pointer,
    other_pointer    : longint;

begin
  abcd:= $abcd;
  idsig:= '!ID!';
  tempstr:= 'TED5v1.0';
  Assign(headfile,GAMEPATH+HEADFILENAME);
  Rewrite(headfile,1);
  Assign(mapfile,GAMEPATH+MAPFILENAME);
  Rewrite(mapfile,1);

  blockwrite(headfile,abcd,2);
  blockwrite(mapfile,tempstr[1],8);
  levelptr:= 8;

  for level:= 1 to LEVELS do
    begin
      with maps[level] do
	begin
	  if GAME_VERSION = 1.1 then
      begin
	map_pointer:= levelptr;
	      s:= Seg(map.data^);
	      o:= Ofs(map.data^);
	      blockwrite(mapfile,Mem[s:o],map.size);
	      Inc(levelptr,map.size);

	object_pointer:= levelptr;
	      s:= Seg(objects.data^);
	      o:= Ofs(objects.data^);
	      blockwrite(mapfile,Mem[s:o],objects.size);
	      Inc(levelptr,objects.size);

	other_pointer:= levelptr;
	      s:= Seg(other.data^);
	      o:= Ofs(other.data^);
	      blockwrite(mapfile,Mem[s:o],other.size);
	      Inc(levelptr,other.size);

	blockwrite(headfile,levelptr,4);

	blockwrite(mapfile,map_pointer,4);
	blockwrite(mapfile,object_pointer,4);
	blockwrite(mapfile,other_pointer,4);
	blockwrite(mapfile,map.size,2);
	blockwrite(mapfile,objects.size,2);
	blockwrite(mapfile,other.size,2);
	blockwrite(mapfile,width,2);
	blockwrite(mapfile,height,2);
	name[0]:=#16;
	blockwrite(mapfile,name[1],16);
	      Inc(levelptr,38);
      end
	  else
	    begin
	      blockwrite(headfile,levelptr,4);
	      map_pointer:= levelptr+38;
	      object_pointer:= map_pointer+map.size;
	      other_pointer:= object_pointer+objects.size;

	blockwrite(mapfile,map_pointer,4);
	blockwrite(mapfile,object_pointer,4);
	blockwrite(mapfile,other_pointer,4);
	blockwrite(mapfile,map.size,2);
	blockwrite(mapfile,objects.size,2);
	blockwrite(mapfile,other.size,2);
	blockwrite(mapfile,width,2);
	blockwrite(mapfile,height,2);
	name[0]:=#16;
	blockwrite(mapfile,name[1],16);

	      s:= Seg(map.data^);
	      o:= Ofs(map.data^);
	      blockwrite(mapfile,Mem[s:o],map.size);
	      s:= Seg(objects.data^);
	      o:= Ofs(objects.data^);
	      blockwrite(mapfile,Mem[s:o],objects.size);
	      s:= Seg(other.data^);
	      o:= Ofs(other.data^);
	      blockwrite(mapfile,Mem[s:o],other.size);
	      Inc(levelptr,map.size+objects.size+other.size+38);
      end;
    blockwrite(mapfile,idsig[1],4);
	  Inc(levelptr,4);
  end;
    end;
  Close(mapfile);
  Close(headfile);
end;

function count_object:integer;
var             i,j                     : integer;
      laskuri : integer;
begin
  laskuri:=0;
  for j:=0 to 63 do
    for i:=0 to 63 do
      if objectmap[i,j]<>0 then
	laskuri:=laskuri+1;
  count_object:=laskuri;
end;

procedure inport(inport_file:string);
var mapfile : file;
    c       : char;
    i,j                     : integer;
    buf                     : array[0..4096] of word;
    search  : searchrec;

begin
  findfirst(inport_file,$ff,search);
  if (doserror=0) or (doserror=3) then
    begin
      Assign(mapfile,inport_file);
      Reset(mapfile,1);
	blockread(mapfile,buf,8192);
	  for j:=0 to 63 do
	    for i:=0 to 63 do
	      levelmap[i,j]:=buf[i+j*64];
	seek(mapfile,8192);
	blockread(mapfile,buf,8192);
	  for j:=0 to 63 do
	    for i:=0 to 63 do
	      objectmap[i,j]:=buf[i+j*64];
      Close(mapfile);
    end
  else
    begin
      SetFillStyle(1,0);
      Bar(5,TEXTLOC,64*7-1+MAP_X,477);
      SetColor(15);
      OutTextXY(30,TEXTLOC,'File '+inport_file+' not found');
      c:=ReadKey;
    end;
end;

procedure export(export_file:string);
var mapfile : file;
    c       : char;
    i,j                     : integer;
    buf              : array[0..4096] of word;
    search  : searchrec;
    (*label skip_write;*)

begin
  findfirst(export_file,$ff,search);
  if (doserror=0) or (doserror=3) then
    begin
      SetFillStyle(1,0);
      Bar(5,TEXTLOC,64*7-1+MAP_X,477);
      SetColor(15);
      OutTextXY(0,TEXTLOC,'File is already exist. Change name or delete old.');
      c:=ReadKey;
      (*goto skip_write;*)
    end
  else
    begin
      Assign(mapfile,export_file);
      Rewrite(mapfile,1);
	for j:=0 to 63 do
	  for i:=0 to 63 do
	    buf[i+j*64]:=levelmap[i,j];
	blockwrite(mapfile,buf,8192);
	for j:=0 to 63 do
	  for i:=0 to 63 do
	    buf[i+j*64]:=objectmap[i,j];
	seek(mapfile,8192);
	blockwrite(mapfile,buf,8192);
      Close(mapfile);
    end;
  (*skip_write:*)
end;

procedure a7a8_expand(src: data_block; var dest: data_block);
var s,o,
    s2,o2,
    index,
    index2,
    size,
    length,
    data,
    newsize  : word;
    goback1  : byte;
    goback2  : word;
    i        : integer;

begin
  s:=Seg(src.data^);
  o:=Ofs(src.data^);
  index:=0;
  Move(Mem[s:o+index],dest.size,2); Inc(index,2);
  GetMem(dest.data,dest.size);
  s2:=Seg(dest.data^);
  o2:=Ofs(dest.data^);
  index2:=0;

  repeat
    Move(Mem[s:o+index],data,2); Inc(index,2);
    case hi(data) of
      $a7: begin
       length:=lo(data);
	     Move(Mem[s:o+index],goback1,1); Inc(index,1);
	     Move(Mem[s2:o2+index2-goback1*2],Mem[s2:o2+index2],length*2);
	     Inc(index2,length*2);
     end;
      $a8: begin
       length:=lo(data);
	     Move(Mem[s:o+index],goback2,2); Inc(index,2);
	     Move(Mem[s2:o2+goback2*2],Mem[s2:o2+index2],length*2);
	     Inc(index2,length*2);
     end;
      else begin
	     Move(data,Mem[s2:o2+index2],2);
	     Inc(index2,2);
     end;
    end;
  until index=src.size;
end;

procedure expand(d: data_block; var g: grid);
var i,x,y : integer;
    s,o,
    data,
    count : word;
    temp  : data_block;
begin
  if GAME_VERSION = 1.1 then
    a7a8_expand(d,temp)
  else
    temp:=d;

  x:= 0;
  y:= 0;
  s:= Seg(temp.data^);
  o:= Ofs(temp.data^);
  Inc(o,2);
  while (y<64) do
    begin
      Move(Mem[s:o],data,2); Inc(o,2);
       if data=$abcd then
	begin
	  Move(Mem[s:o],count,2); Inc(o,2);
	  Move(Mem[s:o],data,2); Inc(o,2);
    for i:= 1 to count do
      begin
	g[x,y]:= data;
	      Inc(x);
	if x=64 then
    begin
      x:= 0;
		  Inc(y);
    end;
      end;
  end
      else
  begin
    g[x,y]:= data;
	  Inc(x);
    if x=64 then
      begin
	x:= 0;
	      Inc(y);
	    end;
	end;
    end;
  if GAME_VERSION = 1.1 then
    FreeMem(temp.data,temp.size);
end;

procedure compress(g: grid; var d: data_block);
var temp     : pointer;
    size: word;
    abcd,
    s,o,
    olddata,
    data,
    nextdata,
    count    : word;
    x,y,i    : integer;
    temp2    : pointer;

begin
  abcd:= $abcd;
  x:= 0;
  y:= 0;
  getmem(temp,8194);
  s:= seg(temp^);
  o:= ofs(temp^);
  data:= $2000;
  move(data,mem[s:o],2);

  size:= 2;
  data:= g[0,0];
  while (y<64) do
    begin
      count:= 1;
      repeat
  inc(x);
  if x=64 then
    begin
      x:=0;
      inc(y);
    end;
  if y<64 then
    nextdata:= g[x,y];
  inc(count);
      until (nextdata<>data) or (y=64);
      dec(count);
      if count<3 then
  begin
    for i:= 1 to count do
      begin
	move(data,mem[s:o+size],2);
	inc(size,2);
      end;
  end
      else
  begin
    move(abcd,mem[s:o+size],2);
    inc(size,2);
    move(count,mem[s:o+size],2);
    inc(size,2);
    move(data,mem[s:o+size],2);
    inc(size,2);
  end;
      data:= nextdata;
    end;
  getmem(temp2,size);
  move(temp^,temp2^,size);
  freemem(temp,8194);
  if GAME_VERSION = 1.1 then
    begin
      getmem(temp,size+2);
      s:= seg(temp^);
      o:= ofs(temp^);
      move(size,mem[s:o],2);
      move(temp2^,mem[s:o+2],size);
      d.data:=temp;
      d.size:= size+2;
      freemem(temp2,size);
    end
  else
    begin
      d.data:= temp2;
      d.size:= size;
    end;
end;

procedure clear_level(n: integer);
var x,y: integer;
begin
   mhide;
   for x:= 0 to 63 do
     for y:= 0 to 63 do
       begin
   levelmap[x,y]:= $8c;
   objectmap[x,y]:= 0;
       end;
   for x:= 0 to 63 do
     begin
       levelmap[x,0]:= 1;
       levelmap[x,63]:= 1;
       levelmap[0,x]:= 1;
       levelmap[63,x]:= 1;
     end;
   display_map;
   mshow;
end;

function str_to_hex(s: string): word;
var temp : word;
    i    : integer;
begin
  temp:= 0;
  for i:= 1 to Length(s) do
    begin
      temp:= temp * 16;
      case s[i] of
	'0'..'9': temp:= temp + Ord(s[i])-Ord('0');
	'a'..'f': temp:= temp + Ord(s[i])-Ord('a')+10;
	'A'..'F': temp:= temp + Ord(s[i])-Ord('A')+10;
      end;
    end;
  str_to_hex:= temp;
end;

procedure showlegend(which,start,n: integer);
var i,x,y: integer;
    save: Boolean;
begin
  mhide;
  save:= show_objects;
  show_objects:= true;
  SetFillStyle(1,0);
  Bar(64*7+MAP_X+13,4,639-5,380-30);
  x:= 66;
  y:= 0;
  for i:= start to start+n-1 do
    begin
      if which=0 then
  begin
    output(x,y,mapgraph[i]);
	  OutText(x+2,y,15,mapnames[i]);
  end
      else
  begin
    output(x,y,objgraph[i]);
	  OutText(x+2,y,15,objnames[i]);
  end;
      Inc(y,2);
    end;
  show_objects:= save;
  mshow;
end;

function inside(x1,y1,x2,y2,x,y: integer): Boolean;
begin
  inside:= (x>=x1) and (x<=x2) and
     (y>=y1) and (y<=y2);
end;

procedure wait_for_mouserelease;
begin
  repeat
    mpos(mouseloc);
  until mouseloc.buttonstatus=0;
end;

procedure bevel(x1,y1,x2,y2,c1,c2,c3: integer);
begin
  SetFillStyle(1,c1);
  Bar(x1,y1,x2,y2);
  SetColor(c2);
  Line(x1,y1,x2,y1);
  Line(x1+1,y1+1,x2-1,y1+1);
  Line(x2,y1,x2,y2);
  Line(x2-1,y1,x2-1,y2-1);
  SetColor(c3);
  Line(x1,y1+1,x1,y2);
  Line(x1+1,y1+2,x1+1,y2);
  Line(x1,y2,x2-1,y2);
  Line(x1+1,y2-1,x2-2,y2-1);
end;

function lue_transfile(s:string):string;
var key         : char;
    enter   :       boolean;
    x                       :       integer;

begin
  enter:=false;
  SetFillStyle(1,0);
  Bar(5,TEXTLOC,64*7-1+MAP_X,477);
  SetColor(15);
  OutTextXY(5,TEXTLOC,'Filename: ');
  x:=1;
  s:='';
  while   not enter do
    begin
      key:=ReadKey;
      if key = #0 then
	key:=readkey;
      if key=#13 then
	enter:=true;
      if key=#8 then          (* BACKSPACE *)
	begin
	  if x>1 then
	    begin
	      x:=x-1;
	      Delete(s,x,x+1);
	      Bar(14*7+x*7,TEXTLOC,15*7+x*7,TEXTLOC+7);
	    end;
	end
      else if (key<>#13) and (key<>#8) then
	begin
	  OutTextXY(14*7+x*7,TEXTLOC,key);
	  Insert(key,s , x);
	  x:=x+1;
	end;
    end;
    lue_transfile:=s;
end;

procedure initialize;
var i: integer;
    infile: text;

    path : pathstr;
    dir  : dirstr;
    name : namestr;
    ext  : extstr;
    filename  : string;
    hexstr    : string[4];
    graphstr  : string[4];
    name20    : string[20];
    junk      : char;
    search    : searchrec;

begin
  filename:= GAMEPATH + HEADFILENAME + '.*';
  WriteLn('searching for ',filename);
  FindFirst(filename,$ff,search);
  if DosError<>0 then
    begin
      WriteLn('Error opening ',HEADFILENAME,' file.');
      WriteLn;
      WriteLn('Be sure that you installed MAPEDIT in the directory where');
      WriteLn('Wolfenstein 3-D is installed.');
      Halt(0);
    end
  else
    begin
      filename:= search.name;
      FSplit(filename,dir,name,ext);
      HEADFILENAME:= upper(HEADFILENAME+ext);
      EXTT:=upper(ext);
      if upper(ext)='.SOD' then
    LEVELS:=21;
      if upper(ext)='.WL1' then
    LEVELS:=10;
      if (upper(ext)='.WL1') or (upper(ext)='.SOD') then
  begin
    GAME_VERSION:=1.0;
    MAPFILENAME:='MAPTEMP'+ext;
    filename:=GAMEPATH+'MAPTEMP'+ext;
	  FindFirst(filename,$ff,search);
	  if DosError<>0 then
      begin
	GAME_VERSION:=1.1;
	MAPFILENAME:='GAMEMAPS'+ext;
	filename:=GAMEPATH+'GAMEMAPS'+ext;
	      FindFirst(filename,$ff,search);
	      if DosError<>0 then
    begin
		  WriteLn('Error opening GAMEMAPS or MAPTEMP file.');
		  Halt(0);
    end;
      end;
  end;
      if (upper(ext)='.WL3') or (upper(ext)='.WL6') then
  begin
    GAME_VERSION:=1.1;
    if upper(ext)='.WL3' then
      LEVELS:= 30
    else
	    LEVELS:= 60;
    MAPFILENAME:='GAMEMAPS'+ext;
    filename:=GAMEPATH+'GAMEMAPS'+ext;
    FindFirst(filename,$ff,search);
    if DosError<>0 then
      begin
	      WriteLn('Error opening GAMEMAPS file.');
	      Halt(0);
      end;
  end;
    end;

  for i:= 0 to 511 do
    begin
      mapnames[i]:= 'unknown '+hex(i);
      objnames[i]:= 'unknown '+hex(i);
      mapgraph[i]:= 'f010';
      objgraph[i]:= 'f010';
    end;
  Assign(infile,'mapdata.def');
  Reset(infile);
  while not Eof(infile) do
    begin
      ReadLn(infile,hexstr,junk,graphstr,junk,name20);
      mapnames[str_to_hex(hexstr)]:= name20;
      mapgraph[str_to_hex(hexstr)]:= graphstr;
    end;
  Close(infile);

  Assign(infile,'objdata.def');
  Reset(infile);
  while not Eof(infile) do
    begin
      ReadLn(infile,hexstr,junk,graphstr,junk,name20);
      objnames[str_to_hex(hexstr)]:= name20;
      objgraph[str_to_hex(hexstr)]:= graphstr;
    end;
  Close(infile);

end;


var gd,gm,
    i,j,x,y,k : integer;
    infile    : text;
    level     : word;
    oldx,oldy : integer;
    done      : Boolean;
    outstr,
    tempstr   : string;
    transmap  : string;
    transhead : string;
    transfile : string;
    transext        : string;

    path : pathstr;
    dir  : dirstr;
    name : namestr;
    ext  : extstr;

    legendpos : integer;
    legendtype: integer;
    newj      : integer;
    currenttype,
    currentval: integer;
    objects         : integer;

    oldj,oldi : integer;

    key       : char;
    control   : Boolean;

begin
  clrscr;
  initialize;
  directvideo:=false;
  read_levels;

  gd:= vga;
  gm:= vgahi;
  initgraph(gd,gm,'');

  settextstyle(0,0,1);
  mreset(themouse);

  show_objects:= true;
  show_floor:= false;

  x:= port[$3da];
  port[$3c0]:= 0;
  SetFillStyle(1,7);
  Bar(0,0,64*7+MAP_X+4,64*7+MAP_Y+4);
  Bar(64*7+MAP_X+9,0,639,380);
  SetFillStyle(1,0);
  Bar(2,2,64*7+MAP_X+2,64*7+MAP_Y+2);
  Bar(64*7+MAP_X+11,2,637,380-28);
  Bar(64*7+MAP_X+11,380-25,637,378);
  SetColor(15);
  OutTextXY(64*7+MAP_X+15,380-16,' MAP  OBJ  UP  DOWN');
  SetFillStyle(1,7);
  Bar(64*7+MAP_X+11+043,380-25,64*7+MAP_X+11+044,378);
  Bar(64*7+MAP_X+11+083,380-25,64*7+MAP_X+11+084,378);
  Bar(64*7+MAP_X+11+113,380-25,64*7+MAP_X+11+114,378);

  legendpos:= 0;
  legendtype:= 0;
  currenttype:= 0;
  currentval:= 1;
  SetFillStyle(1,0);
  Bar(66*7+MAP_X,58*7+MAP_Y,637,61*7+MAP_Y);
  if currenttype=0 then
    begin
      output(66,60,mapgraph[currentval]);
      OutText(67,60,15,' - '+mapnames[currentval]);
    end
  else
    begin
      output(66,60,objgraph[currentval]);
      OutText(67,60,15,' - '+objnames[currentval]);
    end;

  showlegend(legendtype,legendpos,25);

  x:= port[$3da];
  port[$3c0]:= 32;
  mshow;
  level:=1;
  done:= false;
  repeat
    mhide;
    SetFillStyle(1,0);
    Bar(5,TEXTLOC,64*7-1+MAP_X,477);
    SetColor(15);
    OutTextXY(5,TEXTLOC,maps[level].name);
    expand(maps[level].map,levelmap);
    expand(maps[level].objects,objectmap);
    display_map;

		SetFillStyle(1,0);
		Bar(66*7+MAP_X,58*7+MAP_Y,637,59*7+MAP_Y);
		outstr:= 'Objects(amount)=';
		Str(count_object:5,tempstr);
		outstr:= outstr+tempstr;
		SetColor(15);
		OutTextXY(66*7+MAP_X,58*7+MAP_Y,outstr);

    mshow;
    oldx:= 0;
    oldy:= 0;
    key:= #0;
    repeat
      repeat
  mpos(mouseloc);
  x:= mouseloc.column;
  y:= mouseloc.row;
			until (oldx<>x) or (oldy<>y) or KeyPressed or (mouseloc.buttonstatus<>0);
      oldx:= x;
      oldy:= y;
      if (mouseloc.buttonstatus<>0) then
  begin
    if inside(MAP_X,MAP_Y,64*7+MAP_X-1,64*7+MAP_Y-1,x,y) then
      begin
	mhide;
	repeat
    i:= (x - MAP_X) div 7;
    j:= (y - MAP_Y) div 7;
    if currenttype=0 then
      levelmap[i,j]:= currentval
    else
      objectmap[i,j]:= currentval;
		SetFillStyle(1,0);
    dobar(i*7,j*7,i*7+6,j*7+6);
    if show_floor then
      output(i,j,mapgraph[levelmap[i,j]])
    else
      if not (levelmap[i,j] in [$6a..$8f]) then
	output(i,j,mapgraph[levelmap[i,j]]);
    if show_objects then
      output(i,j,objgraph[objectmap[i,j]]);
		mpos(mouseloc);
    x:= mouseloc.column;
    y:= mouseloc.row;
	until (not inside(MAP_X,MAP_Y,64*7+MAP_X-1,64*7+MAP_Y-1,x,y)) or
		    (mouseloc.buttonstatus=0);

	      SetFillStyle(1,0);
	      Bar(66*7+MAP_X,58*7+MAP_Y,637,59*7+MAP_Y);
	      outstr:= 'Objects(amount)=';
	      Str(count_object:5,tempstr);
	      outstr:= outstr+tempstr;
	      SetColor(15);
	      OutTextXY(66*7+MAP_X,58*7+MAP_Y,outstr);

	      mshow;
	    end;
    if inside(464,355,506,378,x,y) then
      begin
	wait_for_mouserelease;
	legendpos:= 0;
	legendtype:= 0;
	showlegend(legendtype,legendpos,25);
      end;
    if inside(509,355,546,378,x,y) then
      begin
	wait_for_mouserelease;
	legendpos:= 0;
	legendtype:= 1;
	showlegend(legendtype,legendpos,25);
      end;
    if inside(549,355,576,378,x,y) then
      begin
	wait_for_mouserelease;
	      Dec(legendpos,25);
	if legendpos<0 then legendpos:= 0;
	showlegend(legendtype,legendpos,25);
      end;
    if inside(579,355,637,378,x,y) then
      begin
	wait_for_mouserelease;
	      Inc(legendpos,25);
	if (legendpos+25)>279 then legendpos:= 279-25;
	showlegend(legendtype,legendpos,25);
      end;
  end;
      if inside(464,2,637,350,x,y) then
  begin
	  mhide;
    j:= (y-2) div 14;
	  SetColor(15);
	  Rectangle(465,j*14+2+1,636,j*14+2+12);
    repeat
      mpos(mouseloc);
      newj:= (mouseloc.row-2) div 14;
      if mouseloc.buttonstatus<>0 then
	begin
    currenttype:= legendtype;
    currentval:= legendpos+j;
		SetFillStyle(1,0);
		Bar(66*7+MAP_X,60*7+MAP_Y,637,61*7+MAP_Y);
    if currenttype=0 then
      begin
	output(66,60,mapgraph[currentval]);
		    OutText(67,60,15,' - '+mapnames[currentval]);
      end
    else
      begin
	output(66,60,objgraph[currentval]);
		    OutText(67,60,15,' - '+objnames[currentval]);
      end;
	end;
	  until (newj<>j) or (mouseloc.column<464) or KeyPressed;
	  SetColor(0);
    Rectangle(465,j*14+2+1,636,j*14+2+12);
    mshow;
  end;

      if inside(MAP_X,MAP_Y,64*7+MAP_X-1,64*7+MAP_Y-1,x,y) then
  begin
    i:= (x - MAP_X) div 7;
    j:= (y - MAP_Y) div 7;
	  if (oldj<>j) or (oldi<>i) then
      begin
	outstr:= '(';
	      Str(i:2,tempstr);
	      outstr:= outstr+tempstr+',';
	      Str(j:2,tempstr);
	      outstr:= outstr+tempstr+')    map: '+hex(levelmap[i,j]);
	      outstr:= outstr+' - '+mapnames[levelmap[i,j]];
	      SetFillStyle(1,0);
	      SetColor(15);
	      Bar(100,TEXTLOC,64*7+MAP_X-1,479);
	      OutTextXY(100,TEXTLOC,outstr);
	outstr:= '        object: '+hex(objectmap[i,j])+' - '+objnames[objectmap[i,j]];
	      OutTextXY(100,TEXTLOC+10,outstr);
	oldj:= j;
	oldi:= i;
      end;
  end
      else
  begin
    mhide;
	  SetFillStyle(1,0);
	  Bar(100,TEXTLOC,440,479);
	  mshow;
	end;

      if KeyPressed then
  begin
	  control:= false;
	  key:= ReadKey;
    if key=#0 then
      begin
	control:= true;
	      key:= ReadKey;
	    end;
    if control then
      case key of
	'H':
    begin
		  FreeMem(maps[level].map.data,maps[level].map.size);
		  FreeMem(maps[level].objects.data,maps[level].objects.size);
      compress(levelmap,maps[level].map);
      compress(objectmap,maps[level].objects);
		  Inc(level);
    end;
	'P':
    begin
		  FreeMem(maps[level].map.data,maps[level].map.size);
		  FreeMem(maps[level].objects.data,maps[level].objects.size);
      compress(levelmap,maps[level].map);
      compress(objectmap,maps[level].objects);
		  Dec(level);
    end;
      end
    else
      case key of
	      'e','E':
		begin
		  mhide;
		  SetFillStyle(1,0);
		  Bar(5,TEXTLOC,64*7-1+MAP_X,477);
		  SetColor(15);
									transfile:=lue_transfile(transfile);
		  FSplit(transfile,dir,name,ext);
		  if ext='' then ext:=EXTT;
		  transfile:=upper(name)+upper(ext);

		  if transfile<>'' then
		    export(transfile);

		  display_map;
		  SetFillStyle(1,0);
		  Bar(0,TEXTLOC,64*7-1+MAP_X,477);
		  SetColor(15);
		  OutTextXY(5,TEXTLOC,maps[level].name);
		  mshow;
		end;
	      'i','I':
		begin
		  mhide;
		  Bar(0,TEXTLOC,64*7-1+MAP_X,477);
		  SetColor(15);
		  transfile:=lue_transfile(transfile);
		  FSplit(transfile,dir,name,ext);
		  if ext='' then ext:=EXTT;
		  transfile:=upper(name)+upper(ext);

		  if transfile<>'' then
		    inport(transfile);

		  display_map;
		  SetFillStyle(1,0);
		  Bar(0,TEXTLOC,64*7-1+MAP_X,477);
		  SetColor(15);
		  OutTextXY(5,TEXTLOC,maps[level].name);
		  mshow;
		end;
	'q','Q':
		begin
		  done:= true;
		  FreeMem(maps[level].map.data,maps[level].map.size);
		  FreeMem(maps[level].objects.data,maps[level].objects.size);
		  compress(levelmap,maps[level].map);
		  compress(objectmap,maps[level].objects);
		end;
	      'c','C': clear_level(level);
	      'o','O': begin
			 mhide;
			 show_objects:= not show_objects;
			 display_map;
       mshow;
	   end;
	'f','F': begin
       mhide;
       show_floor:= not show_floor;
       display_map;
			 if legendtype=0 then
	 showlegend(legendtype,legendpos,25);
       mshow;
	   end;
      end;
  end;
    until done or (key in ['P','H']);
    if level=0 then level:=LEVELS;
    if level=(LEVELS+1) then level:=1;
  until done;

  SetFillStyle(1,0);
  Bar(0,TEXTLOC,639,479);
  SetColor(15);
  OutTextXY(0,TEXTLOC,' Save the current levels to disk? (Y/N) ');

  repeat
    repeat until KeyPressed;
    key:= ReadKey;
    if key=#0 then
      begin
	key:= ReadKey;
  key:= #0;
      end;
  until key in ['y','Y','n','N'];

  if key in ['y','Y'] then write_levels;
  textmode(co80);
  WriteLn('MapEdit 4.1 B               Copyright (c) 1992  Bill Kirby');
  WriteLn;
  WriteLn('This program is intended to be for your personal use only.');
  WriteLn('Distribution of any modified maps may be construed as a ');
  WriteLn('copyright violation by Apogee/ID.');
  WriteLn;
end.

